/**
 * @(#)addr.c : allow user space program to access register
 *
 * Copyright (c) 2002 Prolific Inc. All Rights Reserved.
 *
 * This software is the confidential and proprietary information of Prolific
 * Technology, Inc. ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Prolific
 *
 * Modification History:
 *   #000 2002-09-02 jedy    create file
 **/

#include <linux/config.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/poll.h>
#include <linux/ioport.h>
#include <asm/io.h>
#include "addr.h"

#ifdef CONFIG_DEVFS_FS
#include <linux/devfs_fs_kernel.h>
static devfs_handle_t addr_handle;
#endif

#if 0
#define vi_debug(fmt, arg...)  \
    if (debug)  printk(KERN_NOTICE fmt,##arg)
#else
#define vi_debug(fmt, arg...)   /* donothing */
#endif


static int addr_open(struct inode *inode, struct file *filp)
{
    return 0;
}

static int addr_release(struct inode *inode, struct file *filp)
{
    return 0;
}

static ssize_t addr_read(struct file *filp, char *buf, size_t count, loff_t *offp)
{
    return 0;
}

static ssize_t addr_write(struct file *filp, const char *buf, size_t count, loff_t *offp)
{
    return 0;
}

static int addr_ioctl(struct inode *inode, struct file *filp, unsigned int cmd, unsigned long arg)
{
    int ret = 0;
    struct port_t port;

    if (_IOC_DIR(cmd) & _IOC_READ) {
        if (! access_ok(VERIFY_WRITE, (char *)arg, _IOC_SIZE(cmd)))
            return -EFAULT;
    } else if (_IOC_DIR(cmd) & _IOC_WRITE) {
        if (! access_ok(VERIFY_READ, (char *)arg, _IOC_SIZE(cmd)))
            return -EFAULT;
    }

    if (copy_from_user(&port, (int *)arg, sizeof(struct port_t)))
        return -EFAULT;


    switch(cmd) {
    case ADDR_READL:
        vi_debug("readl %08x\n", port.port);
        if (port.port >= PAGE_OFFSET)
            port.value = readl(port.port);
        else
            port.value = -1;
        if (copy_to_user((void *)arg, &port, sizeof(struct port_t)))
            return -EFAULT;
        break;
    case ADDR_READW:
        vi_debug("readw %08x\n", port.port);
        if (port.port >= PAGE_OFFSET)
            port.value = readw(port.port);
        else
            port.value = -1;

        if (copy_to_user((void *)arg, &port, sizeof(struct port_t)))
            return -EFAULT;
        break;
    case ADDR_READB:
        vi_debug("readb %08x\n", port.port);
        if (port.port >= PAGE_OFFSET)
            port.value = readb(port.port);
        else
            port.value = -1;
        if (copy_to_user((void *)arg, &port, sizeof(struct port_t)))
            return -EFAULT;
        break;
    case ADDR_WRITEL:
        vi_debug("writel %08x = %08x\n", port.port, port.value);
        if (port.port >= PAGE_OFFSET)
            writel(port.value, port.port);
        break;

    case ADDR_WRITEW:
        vi_debug("writew %08x = %04x\n", port.port, (port.value & 0xFFFF));
        if (port.port >= PAGE_OFFSET)
            writew((port.value & 0xFFFF), port.port);
        break;

    case ADDR_WRITEB:
        vi_debug("writeb %08x = %02x\n", port.port, (port.value & 0xFF));
        if (port.port >= PAGE_OFFSET)
            writeb((port.value & 0xFF), port.port);
        break;

    default:
        pr_debug("Invalid ioctl %x\n", cmd);
        ret = -ENOTTY;
    }

    return ret;
}


static struct file_operations addr_fops = {
    owner:      THIS_MODULE,
    open:       addr_open,
    release:    addr_release,
    read:       addr_read,
    write:      addr_write,
    ioctl:      addr_ioctl,
};


#define ADDR_CHAR_MAJOR        201
#define DRIVER_DESC     "Prolific addr driver"
#define DRIVER_VERSION  "v0.1"

static int __init addr_init(void)
{
#ifdef CONFIG_DEVFS_FS
    if (devfs_register_chrdev(ADDR_CHAR_MAJOR, "addr", &addr_fops)) {
        printk(KERN_NOTICE "Can't allocate major number %d for JustCom Video Capture Devices\n",
                ADDR_CHAR_MAJOR);
        return -EAGAIN;
    }

    addr_handle = devfs_register(NULL, "addr", DEVFS_FL_DEFAULT,
                        ADDR_CHAR_MAJOR, 0,
                        S_IFCHR | S_IRUGO | S_IWUGO,
                        &addr_fops, NULL);

#else
    if (register_chrdev(ADDR_CHAR_MAJOR, "addr", &addr_fops)) {
        printk(KERN_NOTICE "Can't allocate major number %d for JustCom Video Capture Devices\n",
                ADDR_CHAR_MAJOR);
        return -EAGAIN;
    }

#endif
    pr_info(DRIVER_DESC" "DRIVER_VERSION"\n");
    return 0;
}

static void __exit addr_exit(void)
{
#ifdef CONFIG_DEVFS_FS
    devfs_unregister(addr_handle);
    devfs_unregister_chrdev(ADDR_CHAR_MAJOR, "addr");

#else
    unregister_chrdev(ADDR_CHAR_MAJOR, "addr");
#endif

}

module_init(addr_init);
module_exit(addr_exit);

MODULE_DESCRIPTION(DRIVER_DESC);
MODULE_AUTHOR("Jedy Wei");
MODULE_SUPPORTED_DEVICE("Prolifc writel/readl helper");


MODULE_PARM(debug, "i");
MODULE_PARM_DESC(debug, "Debugging option");

