//////////////////////////////////////////////////////////////////////
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/slab.h>
#include <linux/init.h>
#include <video/plmedia.h>
#include <linux/devfs_fs_kernel.h>
#include <linux/fs.h>
#include <linux/delay.h>
#include <asm/uaccess.h>
#include "vista.h"

//////////////////////////////////////////////////////////////////////
#define		LM_DATE		"20070126"
#define		LM_VERSION	"1.0.0.1"
MODULE_LICENSE("GPL");

//////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////
static void VISTAInitClient(void);
static int VISTASetFormat(PSENSOR pSensor);
static int VISTAGetFormat(PSENSOR pSensor);
static int VISTASetSync(PSENSOR pSensor);
static int VISTAGetSync(PSENSOR pSensor);
static void VISTARestart(PSENSOR pSensor);

static int SensorOpen(struct inode *pinode, struct file *pfile);
static int SeneorRelese(struct inode *pinode, struct file *pfile);
static int SensorIOCTL(struct inode *pinode, struct file *pfile, 
	unsigned int cmd, unsigned long arg);

static const char VISTA_SHORT_NAME[]="VISTA";
static const char VISTA_FULL_NAME[]="VISTA Camera Sensor";
static const char VISTA_VERSION[]="1.0.0.0";

static struct file_operations SensorFOPS=
{
	owner:		THIS_MODULE,
	open:		   SensorOpen,
	release:	   SeneorRelese,
	ioctl:		SensorIOCTL,
};
	
//////////////////////////////////////////////////////////////////////

PVISTADATA     g_pVISTAData;
PSENSOR		   g_pPLSensor;

//////////////////////////////////////////////////////////////////////
static void VISTAInitClient(void)
{
   	printk( "VISTA: VISTAInitClient()\n" );
  
	g_pVISTAData = kmalloc( sizeof(VISTADATA)+sizeof(SENSOR), GFP_KERNEL );
	g_pPLSensor = g_pVISTAData->pSensor =(PSENSOR)(g_pVISTAData+1);
	
	g_pPLSensor->nType	   = TYPE_SENSOR;	
	g_pPLSensor->szShortName   = VISTA_SHORT_NAME;
	g_pPLSensor->szFullName	   = VISTA_FULL_NAME;
   	g_pPLSensor->szVersion	   = VISTA_VERSION;
	g_pPLSensor->Restart	   = VISTARestart;
	g_pPLSensor->SetFormat	   = VISTASetFormat;
	g_pPLSensor->GetFormat	   = VISTAGetFormat;
	g_pPLSensor->SetSync	   = VISTASetSync;
	g_pPLSensor->GetSync	   = VISTAGetSync;
	
	g_pPLSensor->SyncInfo.nHSyncStart	= 0;
	g_pPLSensor->SyncInfo.nVSyncStart	= 0;
	g_pPLSensor->SyncInfo.byITUBT656	= 0;    /* embedded sync */
	g_pPLSensor->SyncInfo.byHRefSelect	= 1;
	g_pPLSensor->SyncInfo.nHSyncPhase	= 0;
	g_pPLSensor->SyncInfo.byHSyncState	= 0;
	g_pPLSensor->SyncInfo.byVSyncState	= 0;
	g_pPLSensor->SyncInfo.byValidEnable	= 0;
	g_pPLSensor->SyncInfo.byValidState	= 0;
	
	g_pPLSensor->Format.nIsVarioPixel	= 0;
	g_pPLSensor->Format.nInterlaced		= 0;

	g_pVISTAData->DevFS = devfs_register (NULL, "sensor", 
		DEVFS_FL_AUTO_DEVNUM, 0, 0,	//	auto major and minor number
		S_IFCHR | S_IWUSR | S_IRUSR, &SensorFOPS, NULL);
}


//////////////////////////////////////////////////////////////////////
static int __init sm_VISTA_init(void)
{
	printk("VISTA: version %s (%s)\n", LM_VERSION, LM_DATE);
	VISTAInitClient();
	return 0;
}


//////////////////////////////////////////////////////////////////////
static void __exit sm_VISTA_exit(void)
{
   	printk( "VISTA: sm_VISTA_exit()\n" );
}

//////////////////////////////////////////////////////////////////////
// Each client has this additional data
//////////////////////////////////////////////////////////////////////
static int SensorOpen(struct inode *pinode, struct file *pfile)
{
	printk( "VISTA: SensorOpen()\n" );
	return 0;
}

//////////////////////////////////////////////////////////////////////
static int SeneorRelese(struct inode *pinode, struct file *pfile)
{
   	printk( "VISTA: SensorOpen()\n" );
	return 0;
}

//////////////////////////////////////////////////////////////////////
static int SensorIOCTL(struct inode *pinode, struct file *pfile, 
	unsigned int cmd, unsigned long arg)
{
   	printk( "VISTA: SensorIOCTL()\n" );
	return 0;
}

//////////////////////////////////////////////////////////////////////
static int VISTASetFormat(PSENSOR pSensor)
{
	unsigned char		byHSTART, byHSTOP, byVSTART, byVSTOP;
	int			bIsNotSXVGA = 1;
    
   	printk( "VISTA: VISTASetFormat(): %d x %d\n", pSensor->Format.nCropWidth, pSensor->Format.nCropHeight );
	
   	pSensor->Format.uHSTART = 0;
   	pSensor->Format.uVSTART = 0;
   	pSensor->Format.nFPS = 30;		
	pSensor->Format.uHSTOP	= pSensor->Format.uHSTART + (pSensor->Format.nCropWidth << bIsNotSXVGA);
	byHSTART = pSensor->Format.uHSTART>>3;
	byHSTOP = pSensor->Format.uHSTOP>>3;
	printk("VISTA: HSTART = 0x%02X (%u), HSTOP = 0x%02X (%u)\n", byHSTART, byHSTART, byHSTOP, byHSTOP);
	pSensor->Format.uVSTOP = pSensor->Format.uVSTART+pSensor->Format.nCropHeight;
	byVSTART = pSensor->Format.uVSTART>>2;
	byVSTOP = pSensor->Format.uVSTOP>>2;
	pSensor->Format.nInterlaced = 0;
	return 0;
}

//////////////////////////////////////////////////////////////////////
static int VISTAGetFormat(PSENSOR pSensor)
{
   printk( "VISTA: VISTAGetFormat()\n" );
	return 0;	
}

//////////////////////////////////////////////////////////////////////
static int VISTASetSync(PSENSOR pSensor)
{
   	printk( "VISTA: VISTASetSync()\n" );

	pSensor->SyncInfo.byITUBT656     = 0;
	pSensor->SyncInfo.byHRefSelect   = 1;
	pSensor->SyncInfo.nHSyncStart    = 0;
	pSensor->SyncInfo.nVSyncStart    = 0;
	pSensor->SyncInfo.nHSyncPhase    = 0;
	pSensor->SyncInfo.byHSyncState   = 0;
	pSensor->SyncInfo.byVSyncState   = 0;
	pSensor->SyncInfo.byValidEnable  = 0;
	pSensor->SyncInfo.byValidState   = 0;
	return 0;
}

//////////////////////////////////////////////////////////////////////
static int VISTAGetSync(PSENSOR pSensor)
{
   	printk( "VISTA: VISTAGetSync()\n" );
	return 0;	
}

//////////////////////////////////////////////////////////////////////
static void VISTARestart(PSENSOR pSensor)
{
   printk( "VISTA: VISTARestart()\n" );
}

//////////////////////////////////////////////////////////////////////
MODULE_AUTHOR("Jon.Lee");
MODULE_DESCRIPTION("Vista driver");
EXPORT_SYMBOL(g_pPLSensor);
module_init(sm_VISTA_init);
module_exit(sm_VISTA_exit);

//////////////////////////////////////////////////////////////////////

